@@ -1,7 +1,7 @@
 ;;; iiimcf.el --- IIIM Client Framework Support on Emacs
 ;;;                   
 
-;; Copyright (C) 2000 MIYASHITA Hisashi
+;; Copyright (C) 2000-2002 MIYASHITA Hisashi
 
 ;; Author: MIYASHITA Hisashi <himi@m17n.org>
 
@@ -36,7 +36,7 @@
 
 ;;; version.
 
-(defconst iiimcf-version "0.4 (Oikedake)")
+(defconst iiimcf-version "0.6 (Amida)")
 
 ;;; error handling.
 
@@ -473,6 +473,28 @@ If impossible, return nil."
 ;;; Emacs Event management
 ;;;
 
+(defun iiimcf-event-loop (keymap)
+    (if (and (fboundp (function dispatch-event))
+	     (fboundp (function next-event)))
+	;; XEmacs
+	(let ((overriding-local-map keymap)
+	      (input-method-function nil))
+	  (dispatch-event (next-event)))
+      ;; Emacs
+      (let (keyseq cmd)
+	(let ((overriding-local-map keymap)
+	      (input-method-function nil))
+	  (setq keyseq (read-key-sequence
+			nil nil nil t)
+		cmd (key-binding keyseq t)))
+	(if (commandp cmd)
+	    (progn
+	      (setq last-command-event (aref keyseq
+					     (1- (length keyseq)))
+		    last-command this-command
+		    this-command cmd)
+	      (call-interactively cmd))))))
+
 ;; overriding-local-map save/restore functions
 (defun iiimcf-set-overriding-local-map
   (val save &optional buffer-local-p)
@@ -489,7 +511,8 @@ If impossible, return nil."
 
 ;; Emacs-XEmacs imcompatibility absorption layer...
 (defun iiimcf-dispatch-emacs-event-internal ()
-  (let (keyseq cmd)
+  (let ((input-method-function nil)
+	keyseq cmd)
     (if (and (fboundp (function dispatch-event))
 	     (fboundp (function next-event)))
 	;; XEmacs
@@ -617,6 +640,128 @@ If impossible, return nil."
    com-id
    (list (list 'iiimp-im-getimvalues-reply com-id im-id))))
 
+;;;
+;;; handle database
+;;;
+
+;; ((<COM-ID> . A-LIST) ...)
+(defvar iiimcf-com-alist nil)
+;; ((<IM-ID> . A-LIST) ...)
+(defvar iiimcf-im-alist nil)
+;; ((<IC-ID> . A-LIST) ...)
+(defvar iiimcf-ic-alist nil)
+
+;;
+;; register
+;;
+
+(defun iiimcf-register-com-id (com-id)
+  (if (null (assq com-id iiimcf-com-alist))
+      (setq iiimcf-com-alist
+	    (cons
+	     (list com-id)
+	     iiimcf-com-alist))))
+
+(defun iiimcf-register-im-id (im-id)
+  (if (null (assoc im-id iiimcf-im-alist))
+      (setq iiimcf-im-alist
+	    (cons
+	     (list im-id)
+	     iiimcf-im-alist))))
+
+(defun iiimcf-register-ic-id (ic-id)
+  (if (null (assoc ic-id iiimcf-ic-alist))
+      (setq iiimcf-ic-alist
+	    (cons
+	     (list ic-id)
+	     iiimcf-ic-alist))))
+
+;;
+;; unregister
+;;
+
+(defun iiimcf-unregister-com-id (com-id)
+  (let ((slot (assq com-id iiimcf-com-alist)))
+    (if slot (setq iiimcf-com-alist (delq slot iiimcf-com-alist)))))
+
+(defun iiimcf-unregister-im-id (im-id)
+  (let ((slot (assoc im-id iiimcf-im-alist)))
+    (if slot (setq iiimcf-im-alist (delq slot iiimcf-im-alist)))))
+
+(defun iiimcf-unregister-ic-id (ic-id)
+  (let ((slot (assoc ic-id iiimcf-ic-alist)))
+    (if slot (setq iiimcf-ic-alist (delq slot iiimcf-ic-alist)))))
+
+;;
+;; get value
+;;
+
+(defsubst iiimcf-com-id-get (com-id key)
+  (cdr (assq key (cdr (assq com-id iiimcf-com-alist)))))
+
+(defsubst iiimcf-im-id-get (im-id key)
+  (cdr (assq key (cdr (assoc im-id iiimcf-im-alist)))))
+
+(defsubst iiimcf-ic-id-get (ic-id key)
+  (cdr (assq key (cdr (assoc ic-id iiimcf-ic-alist)))))
+
+;;
+;; put value
+;;
+
+(defun iiimcf-com-id-put (com-id key val)
+  (let ((slot (assq com-id iiimcf-com-alist))
+	sslot)
+    (if slot
+	(if (setq sslot (assq key (cdr slot)))
+	    (setcdr sslot val)
+	  (setcdr slot (cons (cons key val) (cdr slot))))
+      (error "COM-ID:%S is not registerd." com-id))))
+
+(defun iiimcf-im-id-put (im-id key val)
+  (let ((slot (assoc im-id iiimcf-im-alist))
+	sslot)
+    (if slot
+	(if (setq sslot (assq key (cdr slot)))
+	    (setcdr sslot val)
+	  (setcdr slot (cons (cons key val) (cdr slot))))
+      (error "IM-ID:%S is not registerd." im-id))))
+
+(defun iiimcf-ic-id-put (ic-id key val)
+  (let ((slot (assoc ic-id iiimcf-ic-alist))
+	sslot)
+    (if slot
+	(if (setq sslot (assq key (cdr slot)))
+	    (setcdr sslot val)
+	  (setcdr slot (cons (cons key val) (cdr slot))))
+      (error "IC-ID:%S is not registerd." ic-id))))
+
+;;
+;; map value
+;;
+
+(defun iiimcf-map-all-coms (func)
+  (let ((ll iiimcf-com-alist))
+    (while ll
+      (funcall func (car (car ll)))
+      (setq ll (cdr ll)))))
+
+(defun iiimcf-map-all-ims (func com-id)
+  (let ((ll iiimcf-im-alist)
+	im-id)
+    (while (setq im-id (car (car ll)))
+      (if (eq (iiimp-im-id-to-com-id im-id) com-id)
+	  (funcall func im-id))
+      (setq ll (cdr ll)))))
+
+(defun iiimcf-map-all-ics (func im-id)
+  (let ((ll iiimcf-ic-alist)
+	ic-id)
+    (while (setq ic-id (car (car ll)))
+      (if (equal (iiimp-ic-id-to-im-id ic-id) im-id)
+	  (funcall func ic-id))
+      (setq ll (cdr ll)))))
+
 ;;;
 ;;; Event manager.
 ;;;
@@ -750,17 +895,6 @@ all handlers that match the rest."
 			       (cdr iiimcf-pumped-message-list)))))))
     ret))
 
-(defun iiimcf-im-destroyic-reply-handler-default (mes)
-  (iiimp-send-message
-   'iiimp-im-trigger-notify-reply
-   (iiimp-message-com-id mes)
-   (iiimp-message-im-id mes)
-   (iiimp-message-ic-id mes)))
-(iiimcf-register-handler
- (function iiimcf-im-trigger-notify-handler-default)
- '(iiimp-im-trigger-notify)
- :after)
-
 (defun iiimcf-im-trigger-notify-handler-default (mes)
   (iiimp-send-message
    'iiimp-im-trigger-notify-reply
@@ -962,6 +1096,254 @@ all handlers that match the rest."
  '(iiimp-im-aux-done)
  :after)
 
+;;;
+;;; create/delete handles
+;;;
+
+(defun iiimcf-connect-com (host port)
+  (let ((com-id
+	 (iiimp-create-network-channel host port)))
+    (iiimcf-register-com-id com-id)
+    com-id))
+
+(defun iiimcf-destroy-com (com-id)
+  (iiimcf-map-all-ims
+   (function iiimcf-disconnect-im)
+   com-id)
+  (iiimcf-unregister-com-id com-id)
+  (iiimp-destroy-network-channel com-id)
+  nil)
+
+(defun iiimcf-connect-im (com-id username &optional client-descriptor)
+  (let (mes im-id)
+    (iiimp-send-message
+     'iiimp-im-connect com-id username)
+    (setq mes (iiimcf-message-manager
+	       com-id
+	       (list
+		(list 'iiimp-im-connect-reply com-id))))
+    (setq im-id (iiimp-message-im-id mes))
+    (iiimcf-register-im-id im-id)
+    (if client-descriptor
+	(iiimcf-send-client-descriptor
+	 com-id im-id client-descriptor))
+    im-id))
+
+(defun iiimcf-disconnect-im (im-id)
+  (iiimcf-map-all-ics
+   (function iiimcf-destroy-ic)
+   im-id)
+  (if (iiimp-check-channel-connection
+       (iiimp-im-id-to-com-id im-id))
+      (progn
+	(iiimp-send-message
+	 'iiimp-im-disconnect
+	 (iiimp-im-id-to-com-id im-id)
+	 im-id)
+	(iiimcf-message-manager
+	 (iiimp-im-id-to-com-id im-id)
+	 (list (list 'iiimp-im-disconnect-reply
+		     (iiimp-im-id-to-com-id im-id) im-id)))))
+  (iiimcf-unregister-im-id im-id)
+  nil)
+
+(defun iiimcf-create-ic (im-id attrs)
+  (let ((com-id (iiimp-im-id-to-com-id im-id))
+	mes ic-id)
+    (iiimp-send-message
+     'iiimp-im-createic
+     com-id im-id attrs)
+    (setq mes
+	  (iiimcf-message-manager
+	   com-id
+	   (list
+	    (list 'iiimp-im-createic-reply
+		  com-id im-id)))
+	  ic-id (iiimp-message-ic-id mes))
+    (iiimcf-register-ic-id ic-id)
+    ic-id))
+
+(defun iiimcf-destroy-ic (ic-id)
+  (if (iiimp-check-channel-connection
+       (iiimp-ic-id-to-com-id ic-id))
+      (progn
+	(iiimp-send-message
+	 'iiimp-im-destroyic
+	 (iiimp-ic-id-to-com-id ic-id)
+	 (iiimp-ic-id-to-im-id ic-id)
+	 ic-id)
+	(iiimcf-message-manager
+	 (iiimp-ic-id-to-com-id ic-id)
+	 (list
+	  (list 'iiimp-im-destroyic-reply
+		(iiimp-ic-id-to-com-id ic-id)
+		(iiimp-ic-id-to-im-id ic-id)
+		ic-id)))))
+  (iiimcf-unregister-ic-id ic-id)
+  nil)
+
+;;;
+;;; forward event
+;;;
+
+(defun iiimcf-forward-event (ic-id ev)
+  (let ((iiimev (iiimcf-translate-emacs-event ev)))
+    (if iiimev
+	(progn
+	  (iiimp-send-message
+	   'iiimp-im-forward-event
+	   (iiimp-ic-id-to-com-id ic-id)
+	   (iiimp-ic-id-to-im-id ic-id)
+	   ic-id
+	   (list 'keyevent
+		 (list iiimev)))
+	   ;; I have no idea why Solaris 8 IIIM server
+	   ;; does not respond to IM_FORWARD_EVENT sometimes.
+	  (iiimcf-message-manager
+	   (iiimp-ic-id-to-com-id ic-id)
+	   (list
+	    (list 'iiimp-im-forward-event-reply
+		  (iiimp-ic-id-to-com-id ic-id)
+		  (iiimp-ic-id-to-im-id ic-id)
+		  ic-id)
+	    (list 'iiimp-im-forward-event
+		  (iiimp-ic-id-to-com-id ic-id)
+		  (iiimp-ic-id-to-im-id ic-id)
+		  ic-id)))
+	  nil)
+      ev)))
+
+;;;
+;;; toggle IM(notify trigger)
+;;;
+
+(defsubst iiimcf-trigger-state (ic-id)
+  (iiimcf-ic-id-get ic-id 'trigger-state))
+
+(defun iiimcf-notify-trigger (ic-id flag)
+  (if flag (setq flag t))
+  (if (eq (iiimcf-trigger-state ic-id) flag)
+      nil
+    (iiimp-send-message
+     'iiimp-im-trigger-notify
+     (iiimp-ic-id-to-com-id ic-id)
+     (iiimp-ic-id-to-im-id ic-id)
+     ic-id flag)
+    (iiimcf-message-manager
+     (iiimp-ic-id-to-com-id ic-id)
+     (list
+      (list 'iiimp-im-trigger-notify-reply
+	    (iiimp-ic-id-to-com-id ic-id)
+	    (iiimp-ic-id-to-im-id ic-id)
+	    ic-id)))
+    (iiimcf-ic-id-put ic-id 'trigger-state flag)))
+
+;;;
+;;; maintain preedit
+;;;
+
+(defsubst iiimcf-get-current-preedit-text (ic-id)
+  (cdr (iiimcf-ic-id-get ic-id 'preedit)))
+
+(defsubst iiimcf-get-current-preedit-caret (ic-id)
+  (car (iiimcf-ic-id-get ic-id 'preedit)))
+
+(defsubst iiimcf-get-current-preedit (ic-id)
+  (iiimcf-ic-id-get ic-id 'preedit))
+
+(defun iiimcf-update-preedit (mes)
+  (let* ((ic-id (iiimp-message-ic-id mes))
+	 (preedit-text (iiimcf-get-current-preedit-text
+			ic-id))
+	 (data (iiimp-message-preedit-draw-data mes))
+	 (caret (aref data 0))
+	 (ch-first (aref data 1))
+	 (ch-len (aref data 2))
+	 (contents (aref data 3))
+	 (str (nth 1 contents))
+	 (ch-second (+ ch-first ch-len))
+	 (len (length preedit-text))
+	 head tail)
+    (if (and (>= len ch-first)
+	     (> ch-first 0))
+	(setq head (substring
+		    preedit-text
+		    0 ch-first)))
+    (if (> len ch-second)
+	(setq tail (substring
+		    preedit-text
+		    ch-second)))
+    (setq str (concat head str tail))
+    (iiimcf-ic-id-put
+     ic-id
+     'preedit
+     (cons caret (if (> (length str) 0) str nil)))))
+(iiimcf-register-handler
+ (function iiimcf-update-preedit)
+ '(iiimp-im-preedit-draw)
+ :before)
+
+;;;
+;;; maintain lookup choice.
+;;;
+;; 'lookup-choice-config
+;; 'lookup-choice
+;; 'candidates
+
+(defun iiimcf-set-lookup-choice-config (mes)
+  (let ((ic-id (iiimp-message-ic-id mes)))
+    (iiimcf-ic-id-put
+     ic-id 'lookup-choice-config
+     (iiimp-message-lookup-choice-start-data mes))))
+(iiimcf-register-handler
+ (function iiimcf-set-lookup-choice-config)
+ '(iiimp-im-lookup-choice-start)
+ :before)
+(defun iiimcf-update-lookup-choice (mes)
+  (let* ((ic-id (iiimp-message-ic-id mes))
+	 (lcdd (iiimp-message-lookup-choice-draw-data mes)))
+    (iiimcf-ic-id-put
+     ic-id 'lookup-choice lcdd)
+    (iiimcf-ic-id-put
+     ic-id 'candidates (aref lcdd 3))))
+(iiimcf-register-handler
+ (function iiimcf-update-lookup-choice)
+ '(iiimp-im-lookup-choice-draw)
+ :before)
+(defun iiimcf-update-lookup-choice-process (mes)
+  (let* ((ic-id (iiimp-message-ic-id mes))
+	 (type (iiimp-message-lookup-choice-process-type mes))
+	 (val (iiimp-message-lookup-choice-process-value mes))
+	 (lcdd (iiimcf-ic-id-get ic-id 'lookup-choice))
+	 (lccfg (iiimcf-ic-id-get ic-id 'lookup-choice-config))
+	 (num (aref lccfg 1))
+	 (idx (aref lcdd 2)))
+    (cond ((eq type 'index)
+	   (setq idx (iiimp-message-lookup-choice-process-value mes)))
+	  ((eq val 'next)
+	   (setq idx (+ idx num)))
+	  ((eq val 'prev)
+	   (setq idx (- idx num)))
+	  ((eq val 'first)
+	   (setq idx (aref lcdd 0)))
+	  ((eq val 'last)
+	   (setq idx (- (aref lcdd 1) num -1))))
+    (aset lcdd 2 idx)))
+(iiimcf-register-handler
+ (function iiimcf-update-lookup-choice-process)
+ '(iiimp-im-lookup-choice-process)
+ :before)
+(defun iiimcf-clear-lookup-choice (mes)
+  (let ((ic-id (iiimp-message-ic-id mes)))
+    (iiimcf-ic-id-put ic-id 'lookup-choice nil)
+    (iiimcf-ic-id-put ic-id 'lookup-choice-config nil)))
+(iiimcf-register-handler
+ (function iiimcf-clear-lookup-choice)
+ '(iiimp-im-lookup-choice-start)
+ :before)
+
+;;;
+
 (provide 'iiimcf)
 
 ;; iiimcf.el ends here.
