@@ -1,7 +1,7 @@
 ;;; iiimp.el --- IIIM Protocol Library
 ;;;                   
 
-;; Copyright (C) 2000-2001 MIYASHITA Hisashi
+;; Copyright (C) 2000-2002 MIYASHITA Hisashi
 
 ;; Author: MIYASHITA Hisashi <himi@m17n.org>
 ;;
@@ -51,9 +51,9 @@
 	 (error
 	  "This system have no function to detect coding-system(or its name)."))))
 
-;; (<COM-ID> <COM-ID> ...)
-(defvar iiimp-communication-list nil
-  "IIIMP effective communication list.")
+(defvar iiimp-current-com-id nil)
+(make-variable-buffer-local 'iiimp-current-com-id)
+(put 'iiimp-current-com-id 'permanent-local t)
 
 (defvar iiimp-async-invocation-handler nil)
 (make-variable-buffer-local 'iiimp-async-invocation-handler)
@@ -301,6 +301,17 @@
 
 ;; COM-ID := [<PROCESS> <BUFFER> <HOST> <PORT> <END-POINT>]
 
+(defun iiimp-construct-com-id (proc buf host port)
+  (let ((obuf (current-buffer)))
+    (if (eq obuf buf)
+	(setq iiimp-current-com-id
+	      (vector proc buf host port (point-min)))
+      (prog2
+	  (set-buffer buf)
+	  (setq iiimp-current-com-id
+		(vector proc buf host port (point-min)))
+	(set-buffer obuf)))))
+
 (defsubst iiimp-com-id-process (com-id)
   (aref com-id 0))
 
@@ -320,41 +331,48 @@
   (aset com-id 4 pt)
   com-id)
 
-(defun iiimp-proc-to-com-id (proc)
-  (let ((ll iiimp-communication-list))
-    (while
-	(and ll
-	     (not (iiimp-com-id-process (car ll))))
-      (setq ll (cdr ll)))
-    (car ll)))
-
-;; IM-ID := (COM-ID id._NUMBER_of_IM)
+;; IM-ID := (id._NUMBER_of_IM COM-ID)
 
 (defsubst iiimp-im-id-to-com-id (im-id)
-  (car im-id))
-
-(defsubst iiimp-im-id-to-handle (im-id)
   (nth 1 im-id))
 
-(defsubst iiimp-construct-im-id (com-id im-num)
-  (list com-id im-num))
+(defsubst iiimp-im-id-to-handle (im-id)
+  (car im-id))
 
-;; IC-ID := (COM-ID id._NUMBER_of_IC id._NUMBER_of_IM)
+(defsubst iiimp-construct-im-id (com-id im-num)
+  (list im-num com-id))
+
+;; IC-ID := (id._NUMBER_of_IC id._NUMBER_of_IM COM-ID)
 
 (defsubst iiimp-ic-id-to-com-id (ic-id)
-  (car ic-id))
+  (nth 2 ic-id))
+
+(defsubst iiimp-ic-id-to-im-id (ic-id)
+  (cdr ic-id))
 
 (defsubst iiimp-ic-id-to-handle (ic-id)
-  (nth 1 ic-id))
+  (car ic-id))
 
 (defsubst iiimp-construct-ic-id (com-id im-num ic-num)
-  (list com-id ic-num im-num))
+  (list ic-num im-num com-id))
+
 
 ;;; network I/O
 
 (defsubst iiimp-check-channel-connection (com-id)
   (eq (process-status (iiimp-com-id-process com-id)) 'open))
 
+(defun iiimp-proc-to-com-id (proc)
+  (let ((oldbuf (current-buffer))
+	(nbuf (process-buffer proc)))
+	(if (not (eq oldbuf nbuf))
+	    (set-buffer nbuf)
+	  (prog1
+	      iiimp-current-com-id
+	    (if (not (eq oldbuf nbuf))
+		(set-buffer nbuf)))
+	  iiimp-current-com-id)))
+
 (defun iiimp-network-sentinel (proc stat)
   (let ((com-id (iiimp-proc-to-com-id proc)))
     (if (and com-id
@@ -383,16 +401,8 @@
 
 (defun iiimp-create-network-channel (host port)
   (let* ((buf (generate-new-buffer " *IIIMP*"))
-	 (proc (iiimp-open-network-channel host port buf))
-	 slot)
-    ;; CREATE COM-ID
-    (setq slot (vector proc buf host port
-		       (save-excursion
-			 (set-buffer buf)
-			 (point-min))))
-    (setq iiimp-communication-list
-	  (cons slot iiimp-communication-list))
-    slot))
+	 (proc (iiimp-open-network-channel host port buf)))
+    (iiimp-construct-com-id proc buf host port)))
 
 (defun iiimp-destroy-network-channel (com-id)
   (let* ((proc (iiimp-com-id-process com-id))
@@ -404,8 +414,6 @@
       (setq stat (process-status proc))
       (sleep-for 1))
     (kill-buffer (iiimp-com-id-buffer com-id))
-    (setq iiimp-communication-list
-	  (delq com-id iiimp-communication-list))
     t))
 
 (defun iiimp-async-invocation-handler-1 (beg end len)
@@ -1453,12 +1461,11 @@
 		   packed
 		   (iiimp-make-padding len)))
 	  ((eq 'input-method type)
-	   (error "NOT SUPPORTED YET.")
-	   ;;(setq packed (iiimp-string-pack val))
+	   (setq packed (iiimp-string-pack val)
+		 len (length packed))
 	   (concat (iiimp-attribid-pack '(0 3))
 		   (iiimp-pack-number ((len . 16)))
-		   packed
-		   (iiimp-make-padding len)))
+		   packed))
 	  (t
 	   (error "Unknown IC ATTRIBID:%S" type)))))
 
